import csv
import yaml
import io
import requests

# https://docs.google.com/spreadsheets/d/1PZ11esT6COK7Y0mMZEHifY9hQeREbNSlRbC6dkYtjNg/edit?usp=sharing
doc = "1PZ11esT6COK7Y0mMZEHifY9hQeREbNSlRbC6dkYtjNg"
terms_tab = "Terms"
types_tab = "Types"
licences_tab = "Licences"
terms_csv = 'https://docs.google.com/spreadsheets/d/%s/gviz/tq?tqx=out:csv&sheet=%s' % (doc, terms_tab)
types_csv = 'https://docs.google.com/spreadsheets/d/%s/gviz/tq?tqx=out:csv&sheet=%s' % (doc, types_tab)
licences_csv = 'https://docs.google.com/spreadsheets/d/%s/gviz/tq?tqx=out:csv&sheet=%s' % (doc, licences_tab)

output_yml = "./src/reeco/schema/schema.yml"
licences_yml = "./src/reeco/schema/licences.yml"
component_yml = "./src/reeco/schema/component.yml"
container_yml = "./src/reeco/schema/container.yml"

split_by = {
    'enum':',',
    'xsd-datatype':',',
    'domain':'|',
    'example-values': '[OR]'
}

def dictify(input_csv):
    dictionary = {}
    r = requests.get(input_csv)
    buff = io.StringIO(r.text)
    cr = csv.DictReader(buff)
    position = 0
    for row in cr:
        position = position + 1
        ID = ''
        if 'Scope' in row.keys():
            ID = row['Scope'] + '/'
        if 'Super term' in row.keys() and row['Super term'] != '':
            ID = ID + row['Super term'] + '/'
        if 'Supertype Id' in row.keys() and row['Supertype Id'] != '':
            ID = ID + row['Supertype Id'] + '/'
        if 'Term' in row.keys(): # Terms
            ID = ID + row['Term'] 
        elif 'Type' in row.keys(): # Types
            ID = ID + row['Type']
        elif 'Code' in row.keys(): # Licences
            ID = ID + row['Code']
        else:
            raise ValueError('Bad row: ' + row)
        dictionary[ID] = {}
        dictionary[ID]['_position'] = position
        for key in row:
            k = key.lower().replace(' ','-')
            v = row[key]
            if k in split_by.keys():
                v = v.split(split_by[k])
            if row[key]:
                dictionary[ID][k] = v
    return dictionary

types = dictify(types_csv)
terms = dictify(terms_csv)
licences = dictify(licences_csv)
schema = {}
schema['types'] = types
schema['terms'] = terms
schema['licences'] = licences
yaml.dump(schema,open(output_yml,'w'))

component = []
container = []
for kt in terms.keys():
    # t = { terms[kt]['term']: None }
    #     for ktt in terms[kt].keys():
    #    t[ktt] = terms[kt][ktt]
    term = terms[kt]
    subterms = []
    # Looking for subterms
    for t in terms.keys():
        subterm = terms[t]
        if subterm['scope'] == term['scope'] and 'super-term' in subterm.keys() and subterm['super-term'] == term['term']:
            subterms.append(subterm)
    if len(subterms) > 0:
        term['terms'] = subterms
    if terms[kt]['scope'] == 'Component':
        component.append(term)
    if terms[kt]['scope'] == 'Container':
        container.append(term)
yaml.dump(component,open(component_yml,'w'))
yaml.dump(container,open(container_yml,'w'))
licences2 = []
for l in licences.keys():
    licences2.append(licences[l])
yaml.dump(licences2,open(licences_yml,'w'))



"""
Python library with utility functions for the REECO Framework.
"""

# __init__.py

# Version of the reeco package
__version__ = "0.0.2-1697540418"

from .schema import Schema
from .validator import Validator


import re
import requests
from schema import Schema, And, Use, Or, Optional, Forbidden, SchemaError, Regex
import sys, os
import frontmatter
from reeco import Schema as ReecoSchema

class Validator:
    def __init__(self):
        REECO = ReecoSchema()
        components = list(map( lambda x: x['type'], REECO.components() ) ) + ['Component'] 
        containers = list(map( lambda x: x['type'], REECO.containers() ) ) + ['Container']
        licences = list(map( lambda x: x['code'], REECO.licences() ) )
        ## Reusable validators
        validateURL = And(str, Regex('^http[s]?://.+$'), error='Value must be an HTTP(S) URL' )
        validateID = And(str, Regex('[^\s]+(/[^\s]+(/[^\s]+)?)?') )

        def resolve(data):
            response = requests.head(data)
            if response.status_code == 200:
                return data
            else:
                raise Exception("URL cannot be resolved")

        resolvableURL = And(Use(lambda fff: resolve(fff)), error="URL cannot be resolved.")

        ## Validators for both containers and components
        both = [
            {Or("component-id", "container-id", only_one=True): str}
        ]
        self._containerErrors = [] + both + [
        ## Mandatory items for containers
        # name
        {'name': str},
        {'type': And(lambda v: v in containers, error='Type must be one of: ' + ", ".join(containers) )},
        # container-id
        {'container-id': validateID},
        # funder
        {Optional('funder'): [{
            'name': str,
            'url': And(validateURL, resolvableURL),
            Optional('grant-agreement'): str
        }]},
        # has-part
        {'has-part': And([validateID], error="The activity must be linked to components.")},
        # ro-crate
        {Optional('ro-crate'): list}
        ]
        self._containerWarnings = [
            ## Warning messages for containers
            # name
            {'funder': And([],error="The activity should have at least one funding organisation.")},
            # ro-crate
            {Optional('ro-crate'): list}
        ]

        #print(licences)
        self._componentErrors = [] + both + [
            # component-id
            {'component-id': validateID },
            {Forbidden('container-id'): object },
            # resource
            {Optional('resource'): str},
            # doi https://www.crossref.org/blog/dois-and-matching-regular-expressions/
            # /^10.\d{4,9}/[-._;()/:A-Z0-9]+$/i
            # /^10.1002/[^\s]+$/i
            # /^10.\d{4}/\d+-\d+X?(\d+)\d+<[\d\w]+:[\d\w]*>\d+.\d+.\w+;\d$/i
            # /^10.1021/\w\w\d++$/i
            # /^10.1207/[\w\d]+\&\d+_\d+$/i
            {Optional('doi'): And(str,
                    Or(
                        Regex(r'^10.\d{4,9}/[-._;()/:A-Z0-9]+$', re.I),
                        Regex(r'^10.1002/[^\s]+$', re.I),
                        Regex(r'^10.\d{4}/\d+-\d+X?(\d+)\d+<[\d\w]+:[\d\w]*>\d+.\d+.\w+;\d$', re.I),
                        #Regex(r'^10.1021/\w\w\d++$', re.I), These two don't seem to work
                        #Regex(r'^10.1207/[\w\d]+\&\d+_\d+$', re.I),
                        error="DOI is invalid"
                    ))},
            # name
            {'name': And(str, error="This annotation should include a short name, as a single value")},
            # description
            {'description': And(str, error="This annotation should include a description, as a single paragraph")},
            # type
            {'type': And(lambda v: v in components, error='Type must be one of: ' + ", ".join(components) )},
            # release-date
            {Optional('release-date'): str},
            # release-number
            {Optional('release-number'): str},
            # release-link
            {Optional('release-link'): And(validateURL, resolvableURL)},
            # changelog ### TODO Check file exists
            {Optional('changelog'): str},
            # licence
            {'licence': And([And(str, lambda v: v in licences, error='Licence must be a list of licence codes from the Reeco Annotation Schema reference.' )], error='Specify at least one licence from the REECO Annotation Schema reference.')},
            # image
            {Optional('image'): And(validateURL, resolvableURL)},
            # logo
            {Optional('logo'): And(validateURL, resolvableURL)},
            # demo
            {Optional('demo'): And(validateURL, resolvableURL)},
            # running-instance
            {Optional('running-instance'): And(validateURL, resolvableURL)},
            # contributors
            {Optional('contributors'): list},
            # related-component
            {'related-components': [And({
                # informed-by
                Optional('informed-by'): Or(str, list),
                # use-case
                Optional('use-case'): Or(str, list),
                # story
                Optional('story'): Or(str, list),
                # persona
                Optional('persona'): Or(str, list),
                # documentation
                Optional('documentation'): Or(str, list),
                # evaluated-in
                Optional('evaluated-in'): Or(str, list),
                # extends-software
                Optional('extends'): Or(str, list),
                # reuses-software
                Optional('reuses'): Or(str, list),
                # serves-data
                Optional('serves'): Or(str, list),
                # produces-data
                Optional('produces'): Or(str, list),
                # reused-in
                Optional('reused-in'): Or(str, list),
                # generated-by
                Optional('generated-by'): Or(str, list)
            }, error="Invalid or malformed annotation")]},
            # bibliography
            {Optional('bibliography'): [{
                # published-in
                Optional('published-in'): Or(str, list),
                # main-publication
                Optional('main-publication'): str,
                Optional('publication'): Or(str, list),
                # main-report
                Optional('main-report'):  str,
                # deliverable-document
                Optional('deliverable-document'):  Or(str, list),
            }]}, # ,error="Invalid or malformed annotation")
            # work-package
            {Optional('work-package'): [validateID]},
            # pilot
            {Optional('pilot'): [validateID]},
            # project
            {Optional('project'): validateID}
        ]

        self._componentWarnings = [
            {'bibliography': And(list,error="Components should have an associated bibliography")}
        ]
    
    def _validate(self, annotations, validators):
        errors = []
        #print(validators)
        for attribute in validators:
            try:
                schema = Schema(attribute, ignore_extra_keys=True)
                schema.validate(annotations)
            except Exception as e:
                #print(type(e))
                #print(dir(e))
                #print('args',e.args)
                #print('autos',e.autos)
                #print('code',e.code)
                #print('errors',e.errors)
                errors.append(e)
        return errors
    
    def asComponent(self, annotations):
        return {
            'error': self._validate(annotations, self._componentErrors),
            'warning': self._validate(annotations, self._componentWarnings)
        }
    
    def asContainer(self, annotations):
        return {
            'error': self._validate(annotations, self._containerErrors),
            'warning': self._validate(annotations, self._containerWarnings)
        }
    
    
if __name__ == '__main__':
    if len( sys.argv ) < 2:
        print ("You must set argument!!!")
        exit(1)
    V = Validator()
    print('File: ' + sys.argv[1])
    report = []
    with open(sys.argv[1]) as f:
        try:
            annotations, content = frontmatter.parse(f.read())
            ## Start validation
            if 'component-id' in annotations.keys():
                ### Validate as component
                report = V.asComponent(annotations)
            elif 'container-id' in annotations.keys():
                ### Validate as container
                report = V.asContainer(annotations)
        except Exception as e:
            # Malformed YAML in Markdown
            report = report + [e]
    print(report)


import os
import yaml
import inspect

class Schema:
    
    def __new__(cls, *args, **kwargs):
        # Create a new instance of Schema
        return super().__new__(cls)
    
    def __init__(self):
        # Initialise
        here = os.path.dirname(os.path.abspath(inspect.getfile(Schema)))
        print(here)
        with open(here + "/schema/schema.yml", "r") as stream:
            try:
                self._yaml = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                print(exc)
                sys.exit()

    def subterms(self, term):
        terms = []
        for t in sorted(self._yaml['terms'], key = lambda pos: self._yaml['terms'][pos]['_position']):
            o = self._yaml['terms'][t]
            if o['scope'] == term['scope'] and 'super-term' in o.keys() and term['term'] == o['super-term']:
                terms.append(o)
        return terms

    def termsFor(self, typee):
        isContainer = False
        if typee == 'Container' or typee in map(lambda o:o['type'], self.containers()):
            isContainer = True
        terms = []
        for t in sorted(self._yaml['terms'], key = lambda pos: self._yaml['terms'][pos]['_position']):
            o = self._yaml['terms'][t]
            if 'super-term' not in o.keys() or o['super-term'] == '':
                if o['scope'] == 'Container' and isContainer:
                    terms.append(o)
                if o['scope'] == 'Component' and not isContainer:
                    terms.append(o)
        return terms

    def subtypes(self, typee):
        types = []
        for t in sorted(self._yaml['types'], key = lambda pos: self._yaml['types'][pos]['_position']):
            o = self._yaml['types'][t]
            if 'supertype-id' in o.keys() and typee == o['supertype-id']:
                types.append(o)
        return types

    def allsubtypes(self, typee):
        all = []
        types = self.subtypes(typee)
        all = all + types
        for t in types:
           all = all + self.allsubtypes(t['type'])
        return all
        

    def containers(self):
        return self.allsubtypes('Container')

    def components(self):
        return self.allsubtypes('Component')

    def types(self):
        types = []
        #sorted(student_Dict, key = lambda name: student_Dict[name].age)
        for t in sorted(self._yaml['types'], key = lambda pos: self._yaml['types'][pos]['_position']):
            o = self._yaml['types'][t]
            ks = ['type', 'label', 'supertype-id']
            t = {}
            for k in ks:
                if k in o:
                    t[k] = o[k]
            types.append(t)
        return types

    def terms(self):
        terms = []
        for t in sorted(self._yaml['terms'], key = lambda pos: self._yaml['terms'][pos]['_position']):
            o = self._yaml['terms'][t]
            terms.append(o)
        return terms
    def licences(self):
        licences = []
        for t in sorted(self._yaml['licences'], key = lambda pos: self._yaml['licences'][pos]['_position']):
            o = self._yaml['licences'][t]
            licences.append(o)
        return licences

    def asString(self, item, key):
        if key in item.keys():
            return item[key]
        else:
            return ''

# class MD:
#     def isComponent(self):
#         return True

# if __name__ == '__main__':
    # reeco = Schema()
    #     print(reeco.types())
    # out = eval("{ 'key': 1}")
#     out = eval("Schema()")
#     print(out, type(out))